import functools

import structlog
from sqlalchemy import text
from sqlalchemy.dialects import postgresql
from sqlalchemy.sql.elements import CompilerElement

from orchestrator.db import db

logger = structlog.get_logger(__name__)


def to_sql_string(stmt: CompilerElement) -> str:
    dialect = postgresql.dialect()  # type: ignore
    return str(stmt.compile(dialect=dialect, compile_kwargs={"literal_binds": True}))


@functools.cache
def get_postgres_version() -> int:
    """Returns the Postgres major version as an int."""
    try:
        # The pg_version_num is pg_major_version * 10000 + pg_minor_version
        pg_version_num = int(db.session.scalar(text("show server_version_num")))
        return pg_version_num // 10000
    except ValueError:
        logger.error("Unable to query Postgres version")
        return 0



from collections.abc import Iterable
from itertools import groupby

import structlog
import typer
from more_itertools import first
from sqlalchemy.dialects import postgresql
from sqlalchemy.sql.dml import UpdateBase

from orchestrator.cli.domain_gen_helpers.types import BlockRelationDict
from orchestrator.cli.helpers.input_helpers import _prompt_user_menu
from orchestrator.cli.helpers.print_helpers import noqa_print
from orchestrator.domain.base import ProductBlockModel, SubscriptionModel

logger = structlog.get_logger(__name__)


def sql_compile(sql: UpdateBase) -> str:
    sql_string = str(sql.compile(dialect=postgresql.dialect(), compile_kwargs={"literal_binds": True}))  # type: ignore[no-untyped-call]
    sql_string = sql_string.replace("\n", "")
    logger.debug("generated SQL", sql_string=sql_string)
    return sql_string


def generic_mapper(prop_name: str, model_diffs: dict[str, dict[str, set[str]]]) -> dict[str, set[str]]:
    model_by_resource = [
        (prop_name_value, model_name)
        for model_name, diff in model_diffs.items()
        for prop_name_value in diff.get(prop_name, [])
    ]
    grouped = groupby(model_by_resource, lambda x: x[0])
    return {k: {val[1] for val in v} for k, v in grouped}


def map_create_fixed_inputs(model_diffs: dict[str, dict[str, set[str]]]) -> dict[str, set[str]]:
    return generic_mapper("missing_fixed_inputs_in_db", model_diffs)


def map_delete_fixed_inputs(model_diffs: dict[str, dict[str, set[str]]]) -> dict[str, set[str]]:
    return generic_mapper("missing_fixed_inputs_in_model", model_diffs)


def map_create_resource_type_relations(model_diffs: dict[str, dict[str, set[str]]]) -> dict[str, set[str]]:
    return generic_mapper("missing_resource_types_in_db", model_diffs)


def map_delete_resource_type_relations(model_diffs: dict[str, dict[str, set[str]]]) -> dict[str, set[str]]:
    return generic_mapper("missing_resource_types_in_model", model_diffs)


def map_create_product_to_product_block_relations(model_diffs: dict[str, dict[str, set[str]]]) -> dict[str, set[str]]:
    return generic_mapper("missing_product_blocks_in_db", model_diffs)


def format_block_relation_to_dict(
    model_name: str,
    block_to_find_in_props: str,
    models: dict[str, type[SubscriptionModel]] | dict[str, type[ProductBlockModel]],
    confirm_warnings: bool,
) -> BlockRelationDict:
    model = models[model_name]
    block_props = model._get_depends_on_product_block_types()
    props = {k for k, v in block_props.items() if v.name == block_to_find_in_props}  # type: ignore

    if len(props) > 1 and not confirm_warnings:
        noqa_print("WARNING: Relating a Product Block multiple times is not supported by this migrator!")
        noqa_print(
            "You will need to create your own migration to create a Product Block Instance for each attribute that is related"
        )
        noqa_print(f"Product Block '{block_to_find_in_props}' has been related multiple times to '{model_name}'")
        noqa_print(f"Attributes the block ('{block_to_find_in_props}') has been related with: {', '.join(props)}")
        noqa_print(f"The relation will only be added to the first attribute ('{first(props)}') want to continue?")

        if _prompt_user_menu([("yes", "yes"), ("no", "no")]) == "no":
            typer.echo("Aborted.")
            raise typer.Exit(code=1)

    return BlockRelationDict(name=model_name, attribute_name=first(props))


def map_create_product_block_relations(
    model_diffs: dict[str, dict[str, set[str]]],
    models: dict[str, type[SubscriptionModel]] | dict[str, type[ProductBlockModel]],
    confirm_warnings: bool,
) -> dict[str, list[BlockRelationDict]]:
    data = generic_mapper("missing_product_blocks_in_db", model_diffs)
    return {
        k: [format_block_relation_to_dict(b, k, models, confirm_warnings) for b in blocks] for k, blocks in data.items()
    }


def map_delete_product_block_relations(model_diffs: dict[str, dict[str, set[str]]]) -> dict[str, set[str]]:
    return generic_mapper("missing_product_blocks_in_model", model_diffs)


def get_product_block_names(pbs: list[type[ProductBlockModel]]) -> Iterable[str]:
    return filter(None, (pb.name for pb in pbs))



# Copyright 2019-2025 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Type

from pydantic import SecretStr as PydanticSecretStr
from pydantic_core import MultiHostUrl, Url
from pydantic_settings import BaseSettings

from orchestrator.utils.expose_settings import SecretStr as OrchSecretStr
from orchestrator.utils.expose_settings import SettingsEnvVariablesSchema, SettingsExposedSchema

EXPOSED_ENV_SETTINGS_REGISTRY: Dict[str, Type[BaseSettings]] = {}
MASK = "**********"


def expose_settings(settings_name: str, base_settings: Type[BaseSettings]) -> Type[BaseSettings]:
    """Decorator to register settings classes."""
    EXPOSED_ENV_SETTINGS_REGISTRY[settings_name] = base_settings
    return base_settings


def mask_value(key: str, value: Any) -> Any:
    key_lower = key.lower()
    is_sensitive_key = "secret" in key_lower or "password" in key_lower

    if is_sensitive_key or isinstance(value, (OrchSecretStr, PydanticSecretStr, MultiHostUrl, Url)):
        return MASK

    return value


def get_all_exposed_settings() -> list[SettingsExposedSchema]:
    """Return all registered settings as dicts."""

    def _get_settings_env_variables(base_settings: Type[BaseSettings]) -> list[SettingsEnvVariablesSchema]:
        """Get environment variables from settings."""
        return [
            SettingsEnvVariablesSchema(env_name=key, env_value=mask_value(key, value))
            for key, value in base_settings.model_dump().items()  # type: ignore
        ]

    return [
        SettingsExposedSchema(name=name, variables=_get_settings_env_variables(base_settings))
        for name, base_settings in EXPOSED_ENV_SETTINGS_REGISTRY.items()
    ]



# Copyright 2019-2020 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from uuid import UUID

import structlog
from sqlalchemy import select

from orchestrator.db import ProcessTable, db
from orchestrator.services import processes
from orchestrator.targets import Target
from orchestrator.workflow import ProcessStatus, StepList, done, init, step, workflow
from pydantic_forms.types import State, UUIDstr

logger = structlog.get_logger(__name__)


def get_process_ids_by_process_statuses(process_statuses: list[ProcessStatus], exclude_ids: list[UUID]) -> list:
    return list(
        db.session.scalars(
            select(ProcessTable.process_id).filter(
                ProcessTable.last_status.in_(process_statuses), ProcessTable.process_id.not_in(exclude_ids)
            )
        )
    )


@step("Find waiting workflows")
def find_waiting_workflows(process_id: UUID) -> State:
    created_process_ids = get_process_ids_by_process_statuses([ProcessStatus.CREATED], exclude_ids=[process_id])
    resumed_process_ids = get_process_ids_by_process_statuses([ProcessStatus.RESUMED], exclude_ids=[process_id])
    waiting_process_ids = get_process_ids_by_process_statuses([ProcessStatus.WAITING], exclude_ids=[process_id])

    return {
        "number_of_waiting_processes": len(waiting_process_ids),
        "waiting_process_ids": waiting_process_ids,
        "created_processes_stuck": len(created_process_ids),
        "created_state_process_ids": created_process_ids,
        "resumed_processes_stuck": len(resumed_process_ids),
        "resumed_state_process_ids": resumed_process_ids,
    }


@step("Resume found workflows")
def resume_found_workflows(
    waiting_process_ids: list[UUIDstr],
    resumed_state_process_ids: list[UUIDstr],
) -> State:
    resume_processes = waiting_process_ids + resumed_state_process_ids

    resumed_process_ids: list = []
    for process_id in resume_processes:
        try:
            process = db.session.get(ProcessTable, process_id)
            if not process:
                continue

            # Workaround the commit disable function
            db.session.info["disabled"] = False

            processes.resume_process(process)
            resumed_process_ids.append(process_id)
        except Exception as exc:
            logger.warning("Could not resume process", process_id=process_id, error=str(exc))
        finally:
            # Make sure to turn it on again
            db.session.info["disabled"] = True

    return {
        "number_of_resumed_process_ids": len(resumed_process_ids),
        "resumed_process_ids": resumed_process_ids,
    }


@step("Restart found CREATED workflows")
def restart_created_workflows(created_state_process_ids: list[UUIDstr]) -> State:
    started_process_ids = []
    for process_id in created_state_process_ids:
        try:
            process = db.session.get(ProcessTable, process_id)
            if not process:
                continue

            # Workaround the commit disable function
            db.session.info["disabled"] = False

            processes.restart_process(process)
            started_process_ids.append(process_id)
        except Exception as exc:
            logger.warning("Could not resume process", process_id=process_id, error=str(exc))
        finally:
            # Make sure to turn it on again
            db.session.info["disabled"] = True

    return {
        "number_of_started_process_ids": len(started_process_ids),
        "started_process_ids": started_process_ids,
    }


@workflow(
    "Resume all workflows that are stuck on tasks with the status 'waiting', 'created' or 'resumed'",
    target=Target.SYSTEM,
)
def task_resume_workflows() -> StepList:
    return init >> find_waiting_workflows >> resume_found_workflows >> restart_created_workflows >> done



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections.abc import Iterable
from copy import deepcopy
from datetime import datetime
from http import HTTPStatus
from typing import Any
from uuid import UUID

from dateutil.parser import isoparse
from more_itertools import flatten
from pydantic import BaseModel
from sqlalchemy import Column

from orchestrator.api.error_handling import raise_status
from orchestrator.db import (
    FixedInputTable,
    ProcessStepTable,
    ProcessSubscriptionTable,
    ProductBlockTable,
    ProductTable,
    ResourceTypeTable,
    SubscriptionInstanceTable,
    SubscriptionInstanceValueTable,
    WorkflowTable,
    db,
)
from orchestrator.db.database import BaseModel as DbBaseModel


def validate(cls: type[DbBaseModel], json_dict: dict, is_new_instance: bool = True) -> dict:
    def is_required(v: Column) -> bool:
        return not v.nullable and (not v.server_default or v.primary_key)

    table = cls.__table__  # type: ignore[attr-defined]
    required_columns = {k: v for k, v, *_ in table.columns._collection if is_required(v)}

    required_attributes: Iterable[str] = required_columns.keys()
    if is_new_instance:
        required_attributes = (key for key in required_attributes if not required_columns[key].primary_key)

    missing_attributes = [key for key in required_attributes if key not in json_dict]
    if missing_attributes:
        detail = f"Missing attributes '{', '.join(missing_attributes)}' for {cls.__name__}"
        raise_status(HTTPStatus.BAD_REQUEST, detail=detail)
    return json_dict


def _merge(cls: type[DbBaseModel], d: dict) -> None:
    o = cls(**d)
    db.session.merge(o)
    db.session.commit()


def save(cls: type[DbBaseModel], json_data: BaseModel) -> None:
    try:
        json_dict = transform_json(json_data.model_dump())
        _merge(cls, json_dict)
    except Exception as e:
        raise_status(HTTPStatus.INTERNAL_SERVER_ERROR, str(e))


def create_or_update(cls: type, obj: BaseModel) -> None:
    try:
        json_dict = transform_json(obj.model_dump())
        _merge(cls, json_dict)
    except Exception as e:
        raise_status(HTTPStatus.INTERNAL_SERVER_ERROR, str(e))


def update(cls: type[DbBaseModel], base_model: BaseModel) -> None:
    json_dict = transform_json(base_model.model_dump())
    table = cls.__table__  # type: ignore[attr-defined]
    pk = list({k: v for k, v, *_ in table.columns._collection if v.primary_key}.keys())[0]
    instance = cls.query.filter(cls.__dict__[pk] == json_dict[pk])
    if not instance:
        raise_status(HTTPStatus.NOT_FOUND)
    json_dict = validate(cls, json_dict, is_new_instance=False)
    try:
        _merge(cls, json_dict)
    except Exception as e:
        raise_status(HTTPStatus.INTERNAL_SERVER_ERROR, str(e))


def delete(cls: type[DbBaseModel], primary_key: UUID) -> None:
    table = cls.__table__  # type: ignore[attr-defined]
    pk = list({k: v for k, v, *_ in table.columns._collection if v.primary_key}.keys())[0]
    row_count = cls.query.filter(cls.__dict__[pk] == primary_key).delete()
    db.session.commit()
    if row_count > 0:
        return
    raise_status(HTTPStatus.NOT_FOUND)


deserialization_mapping = {
    "steps": ProcessStepTable,
    "subscriptions": ProcessSubscriptionTable,
    "product_blocks": ProductBlockTable,
    "fixed_inputs": FixedInputTable,
    "resource_types": ResourceTypeTable,
    "instances": SubscriptionInstanceTable,
    "values": SubscriptionInstanceValueTable,
    "products": ProductTable,
    "workflows": WorkflowTable,
}

forbidden_fields = ["created_at"]
date_fields = ["end_date"]


def cleanse_json(json_dict: dict) -> None:
    copy_json_dict = deepcopy(json_dict)
    for k in copy_json_dict.keys():
        if copy_json_dict[k] is None:
            del json_dict[k]
    for forbidden in forbidden_fields:
        if forbidden in json_dict:
            del json_dict[forbidden]
        rel: dict
        for rel in flatten(list(filter(lambda i: isinstance(i, list), json_dict.values()))):
            cleanse_json(rel)


def parse_date_fields(json_dict: dict) -> None:
    for date_field in date_fields:
        if date_field in json_dict:
            val = json_dict[date_field]
            if isinstance(val, float) or isinstance(val, int):
                json_dict[date_field] = datetime.fromtimestamp(val / 1e3)
            if isinstance(val, str):
                timestamp = isoparse(val)
                assert timestamp.tzinfo is not None, "All timestamps should contain timezone information."  # noqa: S101
                json_dict[date_field] = timestamp
        rel: dict
        for rel in flatten(list(filter(lambda i: isinstance(i, list), json_dict.values()))):
            parse_date_fields(rel)


def transform_json(json_dict: dict) -> dict:
    def _contains_list(coll: Iterable[Any]) -> bool:
        return len(list(filter(lambda item: isinstance(item, list), coll))) > 0

    def _do_transform(items: Iterable[tuple[str, Any]]) -> dict:
        return dict(map(_parse, items))

    def _parse(item: tuple[str, Any]) -> tuple[str, Any]:
        if isinstance(item[1], list):
            cls = deserialization_mapping[item[0]]
            return item[0], list(map(lambda i: cls(**_do_transform(i.items())), item[1]))
        return item

    cleanse_json(json_dict)
    parse_date_fields(json_dict)

    if _contains_list(json_dict.values()):
        return _do_transform(json_dict.items())

    return json_dict



# Copyright 2019-2023 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from uuid import UUID

import structlog
from sqlalchemy import BinaryExpression, ColumnElement, false, select

from orchestrator.db import ProcessSubscriptionTable, ProcessTable, ProductTable, SubscriptionTable, WorkflowTable
from orchestrator.db.filters import create_memoized_field_list, generic_filter_from_clauses
from orchestrator.db.filters.search_filters import (
    default_inferred_column_clauses,
    filter_exact,
    inferred_filter,
    node_to_str_val,
)
from orchestrator.db.filters.search_filters.inferred_filter import filter_uuid_exact
from orchestrator.utils.search_query import Node, WhereCondGenerator

logger = structlog.get_logger(__name__)


def make_product_clause(filter_generator: WhereCondGenerator) -> WhereCondGenerator:
    """The passed filter_generator takes a Node and returns a where clause acting on a ProductTable column."""

    def product_clause(node: Node) -> BinaryExpression:
        process_subscriptions = (
            select(ProcessSubscriptionTable.process_id)
            .join(SubscriptionTable)
            .join(ProductTable)
            .where(filter_generator(node))
            .subquery()
        )
        return ProcessTable.process_id.in_(process_subscriptions)

    return product_clause


def customer_clause(node: Node) -> BinaryExpression[bool] | ColumnElement[bool]:
    value = node_to_str_val(node)
    try:
        customer_uuid = str(UUID(value))
    except (ValueError, AttributeError):
        # Not a valid uuid, skip matching with customer_id
        return false()

    process_subscriptions = (
        select(ProcessSubscriptionTable.process_id)
        .join(SubscriptionTable)
        .where(SubscriptionTable.customer_id == customer_uuid)
        .subquery()
    )
    return ProcessTable.process_id.in_(process_subscriptions)


def make_subscription_id_clause(filter_generator: WhereCondGenerator) -> WhereCondGenerator:
    def subscription_id_clause(node: Node) -> BinaryExpression:

        process_subscriptions = select(ProcessSubscriptionTable.process_id).where(filter_generator(node)).subquery()

        return ProcessTable.process_id.in_(process_subscriptions)

    return subscription_id_clause


def make_workflow_field_clause(filter_generator: WhereCondGenerator) -> WhereCondGenerator:
    def workflow_name_clause(node: Node) -> BinaryExpression[bool]:
        process_workflow = (
            select(ProcessTable.process_id)
            .join(WorkflowTable)
            .where(WorkflowTable.deleted_at.is_(None), filter_generator(node))
            .subquery()
        )

        return ProcessTable.process_id.in_(process_workflow)

    return workflow_name_clause


PROCESS_TABLE_COLUMN_CLAUSES = default_inferred_column_clauses(ProcessTable) | {
    "customer": customer_clause,
    "product": make_product_clause(inferred_filter(ProductTable.name)),
    "product_description": make_product_clause(inferred_filter(ProductTable.description)),
    "subscription_id": make_subscription_id_clause(filter_uuid_exact(ProcessSubscriptionTable.subscription_id)),
    "tag": make_product_clause(filter_exact(ProductTable.tag)),
    "target": make_workflow_field_clause(inferred_filter(WorkflowTable.target)),
    "workflow_name": make_workflow_field_clause(inferred_filter(WorkflowTable.name)),
}


process_filter_fields = create_memoized_field_list(PROCESS_TABLE_COLUMN_CLAUSES)
filter_processes = generic_filter_from_clauses(PROCESS_TABLE_COLUMN_CLAUSES)



"""Add generic workflows to core.

Revision ID: a76b9185b334
Revises: 3323bcb934e7
Create Date: 2020-10-19 09:17:49.395612

"""

from uuid import uuid4

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "a76b9185b334"
down_revision = "c112305b07d3"
branch_labels = None
depends_on = None

workflows = [
    {"name": "modify_note", "description": "Modify Note", "workflow_id": uuid4(), "target": "MODIFY"},
    {"name": "task_clean_up_tasks", "description": "Clean up old tasks", "workflow_id": uuid4(), "target": "SYSTEM"},
    {
        "name": "task_resume_workflows",
        "description": "Resume all workflows that are stuck on tasks with the status 'waiting'",
        "workflow_id": uuid4(),
        "target": "SYSTEM",
    },
]


def upgrade() -> None:
    conn = op.get_bind()
    for workflow in workflows:
        conn.execute(
            sa.text(
                "INSERT INTO workflows VALUES (:workflow_id, :name, :target, :description, now()) ON CONFLICT DO NOTHING"
            ),
            workflow,
        )


def downgrade() -> None:
    conn = op.get_bind()
    for workflow in workflows:
        conn.execute(sa.text("DELETE FROM workflows WHERE name = :name"), {"name": workflow["name"]})



"""Add is_task to workflow.

Revision ID: 161918133bec
Revises: 68d14db1b8da
Create Date: 2025-05-08 11:25:51.966410

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "161918133bec"
down_revision = "68d14db1b8da"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("workflows", sa.Column("is_task", sa.Boolean(), server_default=sa.text("false"), nullable=False))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("workflows", "is_task")
    # ### end Alembic commands ###



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pathlib import Path

from pydantic_settings import BaseSettings


class ProductGeneratorSettings(BaseSettings):
    FOLDER_PREFIX: Path = Path("")
    CUSTOM_TEMPLATES: Path = Path("")

    PRODUCT_TYPES_PATH: Path = Path("products/product_types")
    PRODUCT_BLOCKS_PATH: Path = Path("products/product_blocks")
    WORKFLOWS_PATH: Path = Path("workflows")
    TEST_PRODUCT_TYPE_PATH: Path = Path("test/unit_tests/domain/product_types")
    TEST_WORKFLOWS_PATH: Path = Path("test/unit_tests/workflows")

    # Files that will be updated
    PRODUCT_REGISTRY_PATH: Path = Path("products/__init__.py")
    SUBSCRIPTION_DESCRIPTION_PATH: Path = Path("products/services/subscription.py")
    TRANSLATION_PATH: Path = Path("translations/en-GB.json")


product_generator_settings = ProductGeneratorSettings()



"""Add task_validate_products.

Revision ID: 3c8b9185c221
Revises: 3323bcb934e7
Create Date: 2020-04-06 09:17:49.395612

"""

from uuid import uuid4

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "3c8b9185c221"
down_revision = "3323bcb934e7"
branch_labels = None
depends_on = None

workflows = [
    {"name": "task_validate_products", "description": "Validate products", "workflow_id": uuid4(), "target": "SYSTEM"},
]


def upgrade() -> None:
    conn = op.get_bind()
    for workflow in workflows:
        conn.execute(
            sa.text(
                "INSERT INTO workflows VALUES (:workflow_id, :name, :target, :description, now()) ON CONFLICT DO NOTHING"
            ),
            workflow,
        )


def downgrade() -> None:
    conn = op.get_bind()
    for workflow in workflows:
        conn.execute(sa.text("DELETE FROM workflows WHERE name = :name"), {"name": workflow["name"]})



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from collections.abc import Generator
from dataclasses import dataclass
from datetime import datetime
from http import HTTPStatus
from shlex import shlex
from typing import Any
from uuid import UUID

from more_itertools import chunked
from sqlalchemy import Select, func, select
from sqlalchemy.sql import expression
from starlette.responses import Response
from structlog import get_logger

from orchestrator.api.error_handling import raise_status
from orchestrator.db import ProductTable, SubscriptionTable, db
from orchestrator.db.models import SubscriptionSearchView
from orchestrator.db.range.range import Selectable, apply_range_to_statement
from orchestrator.domain.base import SubscriptionModel
from orchestrator.utils.search_query import create_ts_query_string

logger = get_logger(__name__)


def _quote_if_kv_pair(token: str) -> str:
    return f'"{token}"' if ":" in token else token


def _process_text_query(q: str) -> str:
    quote = '"'
    if q.count(quote) % 2 == 1:
        q += quote  # Add missing closing quote
    lex = shlex(q)
    lex.whitespace_split = True
    lex.quotes = '"'
    try:
        return " ".join([_quote_if_kv_pair(token) for token in lex])
    except ValueError:
        logger.debug("Error parsing text query.")
        return q


def _add_sort_to_query(query: Select, sort: list[str] | None) -> Select:
    if sort is None or len(sort) < 2:
        return query

    for item in chunked(sort, 2):
        if item and len(item) == 2:
            order_by_expr = expression.desc if item[1].upper() == "DESC" else expression.asc
            if item[0] in ["product", "tag"]:
                field = "name" if item[0] == "product" else "tag"
                col = ProductTable.__dict__.get(field)
            else:
                col = SubscriptionTable.__dict__.get(item[0])

            if col:
                query = query.order_by(order_by_expr(col))
            else:
                raise_status(HTTPStatus.BAD_REQUEST, f"Unable to sort on unknown field: {item[0]}")
    return query


def add_response_range(
    stmt: Selectable, range_: list[int] | None, response: Response, unit: str = "items"
) -> Selectable:
    if range_ is not None and len(range_) == 2:
        total = db.session.scalar(select(func.count()).select_from(stmt.subquery()))
        range_start, range_end = range_
        try:
            stmt = apply_range_to_statement(stmt, range_start, range_end)
        except ValueError as e:
            logger.exception(e)
            raise_status(HTTPStatus.BAD_REQUEST, str(e))

        response.headers["Content-Range"] = f"{unit} {range_start}-{range_end}/{total}"
    return stmt


MAX_QUERY_STRING_LENGTH = 512


def add_subscription_search_query_filter(stmt: Select, search_query: str) -> Select:
    """Filters the Select statement on the contents of the query string.

    The Select statement should read from SubscriptionTable as a source.
    The query will first be converted from camelCase to snake_case before parsing.
    """
    if len(search_query) > MAX_QUERY_STRING_LENGTH:
        raise_status(HTTPStatus.BAD_REQUEST, f"Max query length of {MAX_QUERY_STRING_LENGTH} characters exceeded.")

    ts_query = create_ts_query_string(search_query)
    return stmt.join(SubscriptionSearchView).filter(
        func.to_tsquery("simple", ts_query).op("@@")(SubscriptionSearchView.tsv)
    )


VALID_SORT_KEYS = {
    "creator": "created_by",
    "started": "started_at",
    "status": "last_status",
    "assignee": "assignee",
    "modified": "last_modified_at",
    "workflow": "workflow",
}


@dataclass
class ProductEnriched:
    product_id: UUID
    description: str
    name: str
    tag: str
    status: str
    product_type: str


@dataclass
class _Subscription:
    customer_id: UUID
    description: str
    end_date: float
    insync: bool
    start_date: float
    status: str
    subscription_id: UUID
    product: ProductEnriched


@dataclass
class _ProcessListItem:
    assignee: str
    created_by: str | None
    failed_reason: str | None
    last_modified_at: datetime
    pid: UUID
    started_at: datetime
    last_status: str
    last_step: str | None
    subscriptions: list[_Subscription]
    workflow: str
    workflow_target: str | None
    is_task: bool


def update_in(dct: dict | list, path: str, value: Any, sep: str = ".") -> None:
    """Update a value in a dict or list based on a path."""
    for x in path.split(sep):
        prev: dict | list
        if x.isdigit() and isinstance(dct, list):
            prev = dct
            dct = dct[int(x)]
        else:
            prev = dct
            dct = dict(dct).setdefault(x, {})
    prev[x] = value  # type: ignore


def get_in(dct: dict | list, path: str, sep: str = ".") -> Any:
    """Get a value in a dict or list using the path and get the resulting key's value."""
    prev: dict | list
    for x in path.split(sep):
        if x.isdigit() and isinstance(dct, list):
            prev, dct = dct, dct[int(x)]
        else:
            prev, dct = dct, dict(dct).get(x)  # type: ignore
    return prev[x]  # type: ignore


def getattr_in(obj: Any, attr: str) -> Any:
    """Get an instance attribute value by path."""

    def _getattr(obj: object, attr: str) -> Any:
        if isinstance(obj, list):
            return obj[int(attr)]

        if isinstance(obj, dict):
            return obj.get(attr)

        return getattr(obj, attr, None)

    return functools.reduce(_getattr, [obj] + attr.split("."))


def product_block_paths(subscription: SubscriptionModel | dict) -> list[str]:
    _subscription = subscription.model_dump() if isinstance(subscription, SubscriptionModel) else subscription

    def get_dict_items(d: dict) -> Generator:
        for k, v in d.items():
            if isinstance(v, dict):
                for k1, v1 in get_dict_items(v):
                    yield (f"{k}.{k1}", v1)
                yield (k, v)
            if isinstance(v, list):
                for index, list_item in enumerate(v):
                    if isinstance(list_item, dict):
                        for list_item_key, list_item_value in get_dict_items(list_item):
                            yield (f"{k}.{index}.{list_item_key}", list_item_value)
                        yield (f"{k}.{index}", list_item)

    return [path for path, value in get_dict_items(_subscription)]



import structlog
from sqlalchemy import func, select

from orchestrator.db import db
from orchestrator.db.filters import Filter
from orchestrator.db.filters.product import PRODUCT_TABLE_COLUMN_CLAUSES, filter_products, product_filter_fields
from orchestrator.db.models import ProductTable
from orchestrator.db.range.range import apply_range_to_statement
from orchestrator.db.sorting import Sort
from orchestrator.db.sorting.product import product_sort_fields, sort_products
from orchestrator.graphql.pagination import Connection
from orchestrator.graphql.resolvers.helpers import rows_from_statement
from orchestrator.graphql.schemas.product import ProductType
from orchestrator.graphql.types import GraphqlFilter, GraphqlSort, OrchestratorInfo
from orchestrator.graphql.utils import create_resolver_error_handler, is_querying_page_data, to_graphql_result_page
from orchestrator.graphql.utils.get_query_loaders import get_query_loaders_for_gql_fields
from orchestrator.utils.search_query import create_sqlalchemy_select

logger = structlog.get_logger(__name__)


async def resolve_products(
    info: OrchestratorInfo,
    filter_by: list[GraphqlFilter] | None = None,
    sort_by: list[GraphqlSort] | None = None,
    first: int = 10,
    after: int = 0,
    query: str | None = None,
) -> Connection[ProductType]:
    _error_handler = create_resolver_error_handler(info)

    pydantic_filter_by: list[Filter] = [item.to_pydantic() for item in filter_by] if filter_by else []
    pydantic_sort_by: list[Sort] = [item.to_pydantic() for item in sort_by] if sort_by else []
    logger.debug("resolve_products() called", range=[after, after + first], sort=sort_by, filter=pydantic_filter_by)

    query_loaders = get_query_loaders_for_gql_fields(ProductTable, info)
    select_stmt = select(ProductTable).options(*query_loaders)
    select_stmt = filter_products(select_stmt, pydantic_filter_by, _error_handler)

    if query is not None:
        stmt = create_sqlalchemy_select(
            select_stmt,
            query,
            mappings=PRODUCT_TABLE_COLUMN_CLAUSES,
            base_table=ProductTable,
            join_key=ProductTable.product_id,
        )
    else:
        stmt = select_stmt

    stmt = sort_products(stmt, pydantic_sort_by, _error_handler)
    total = db.session.scalar(select(func.count()).select_from(stmt.subquery()))
    stmt = apply_range_to_statement(stmt, after, after + first + 1)

    graphql_products = []
    if is_querying_page_data(info):
        products = rows_from_statement(stmt, ProductTable, unique=True, loaders=query_loaders)
        graphql_products = [ProductType.from_pydantic(p) for p in products]
    return to_graphql_result_page(graphql_products, first, after, total, product_sort_fields(), product_filter_fields())



from collections.abc import Callable

from more_itertools import flatten, one
from strawberry.types.nodes import FragmentSpread, InlineFragment, SelectedField, Selection

from orchestrator.graphql.types import OrchestratorInfo


def get_selected_field_selections(selected_field: Selection, name: str) -> list[Selection]:
    def is_named_field(selection: Selection, field_name: str) -> bool:
        return isinstance(selection, SelectedField) and selection.name == field_name

    page_field = [selection for selection in selected_field.selections if is_named_field(selection, name)]
    return one(page_field).selections if page_field else []


def is_query_detailed(basic_fields: tuple, type_conditions: tuple | None = None) -> Callable[[OrchestratorInfo], bool]:
    """Function wrapper that validates GraphQL queries against provided basic_fields.

    Args:
        basic_fields: tuple of the basic fields.
        type_conditions: optional tuple of the type conditions.

    Returns:
        Function that validates if a GraphQL query includes fields beyond the specified basic_fields.
    """

    def _is_model_info_detailed(info: OrchestratorInfo) -> bool:
        """Check if the query asks for detailed props."""

        def has_details(selection: Selection) -> bool:
            if isinstance(selection, InlineFragment):
                if type_conditions and selection.type_condition not in type_conditions:
                    return True
                return any(has_details(selection) for selection in selection.selections)
            if isinstance(selection, FragmentSpread):
                return any(has_details(s) for s in selection.selections)
            return selection.name not in basic_fields

        fields = flatten(
            (get_selected_field_selections(field, "page") or field.selections) for field in info.selected_fields
        )
        return any(has_details(selection) for selection in fields if selection)

    return _is_model_info_detailed


def is_querying_page_data(info: OrchestratorInfo) -> bool:
    return any(flatten(get_selected_field_selections(field, "page") for field in info.selected_fields))



# Copyright 2019-2025 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions to transform result of query SubscriptionInstanceAsJsonFunction to match the ProductBlockModel."""

from functools import partial
from typing import TYPE_CHECKING, Any, Callable, Iterable

from more_itertools import first, only

from orchestrator.types import is_list_type, is_optional_type

if TYPE_CHECKING:
    from orchestrator.domain.base import ProductBlockModel


def _ensure_list(instance_or_value_list: Any) -> Any:
    if instance_or_value_list is None:
        return []

    return instance_or_value_list


def _instance_list_to_dict(product_block_field_type: type, instance_list: Any) -> Any:
    if instance_list is None:
        return None

    match instance_list:
        case list():
            if instance := only(instance_list):
                return instance

            if not is_optional_type(product_block_field_type):
                raise ValueError("Required subscription instance is missing in database")

            return None  # Set the optional product block field to None
        case _:
            raise ValueError(f"All subscription instances should be returned as list, found {type(instance_list)}")  #


def _value_list_to_value(field_type: type, value_list: Any) -> Any:
    if value_list is None:
        return None

    match value_list:
        case list():
            if (value := only(value_list)) is not None:
                return value

            if not is_optional_type(field_type):
                raise ValueError("Required subscription value is missing in database")

            return None  # Set the optional resource type field to None
        case _:
            raise ValueError(f"All instance values should be returned as list, found {type(value_list)}")


def field_transformation_rules(klass: type["ProductBlockModel"]) -> dict[str, Callable]:
    """Create mapping of transformation rules for the given product block type."""

    def create_rules() -> Iterable[tuple[str, Callable]]:
        for field_name, product_block_field_type in klass._product_block_fields_.items():
            if is_list_type(product_block_field_type):
                yield field_name, _ensure_list
            else:
                yield field_name, partial(_instance_list_to_dict, product_block_field_type)

        for field_name, field_type in klass._non_product_block_fields_.items():
            if is_list_type(field_type):
                yield field_name, _ensure_list
            else:
                yield field_name, partial(_value_list_to_value, field_type)

    return dict(create_rules())


def transform_instance_fields(all_rules: dict[str, dict[str, Callable]], instance: dict) -> None:
    """Apply transformation rules to the given subscription instance dict."""

    from orchestrator.domain.base import ProductBlockModel

    # Lookup applicable rules through product block name
    field_rules = all_rules[instance["name"]]

    klass = ProductBlockModel.registry[instance["name"]]

    # Ensure the product block's metadata is loaded
    klass._fix_pb_data()

    # Transform all fields in this subscription instance
    try:
        for field_name, rewrite_func in field_rules.items():
            field_value = instance.get(field_name)
            instance[field_name] = rewrite_func(field_value)
    except ValueError as e:
        raise ValueError(f"Invalid subscription instance data {instance}") from e

    # Recurse into nested subscription instances
    for field_value in instance.values():
        if isinstance(field_value, dict):
            transform_instance_fields(all_rules, field_value)
        if isinstance(field_value, list) and isinstance(first(field_value, None), dict):
            for list_value in field_value:
                transform_instance_fields(all_rules, list_value)



# Copyright 2019-2023 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
from collections.abc import Callable, Iterable
from functools import cache
from typing import Any, Protocol

from more_itertools import partition
from pydantic import BaseModel
from sqlalchemy import Select, false

from orchestrator.utils.helpers import camel_to_snake, to_camel
from orchestrator.utils.search_query import Node, SQLAlchemyVisitor, WhereCondGenerator


class CallableErrorHandler(Protocol):
    def __call__(self, message: str, **kwargs: Any) -> None: ...


class Filter(BaseModel):
    field: str
    value: str


QueryType = Select
ValidFilterFunctionsByColumnType = dict[str, Callable[[QueryType, str], QueryType]]
ColumnMappings = dict[str, WhereCondGenerator]


def generic_filters_validate(
    valid_columns: Iterable[str],
) -> Callable[[list[Filter]], tuple[Iterable[Filter], Iterable[Filter]]]:
    def _validate_filters(filter_by: list[Filter]) -> tuple[Iterable[Filter], Iterable[Filter]]:
        def _is_valid_filter(item: Filter) -> bool:
            return camel_to_snake(item.field) in valid_columns and item.value is not None

        return partition(_is_valid_filter, filter_by)

    return _validate_filters


_re_split = re.compile("[-|]")


def _filter_to_node(filter_item: Filter) -> Node:
    value = filter_item.value
    if value and value[0] == "!":
        updated_filter_item = Filter(field=filter_item.field, value=value[1:])
        return "Negation", _filter_to_node(updated_filter_item)

    # Workaround to deal with date and id fields. These should not be split by '-' like other fields.
    # Fix after deprecating '-' as split-separator in favor of '|'
    should_split = "date" not in filter_item.field.lower() and not filter_item.field.lower().endswith("id")
    values = _re_split.split(filter_item.value) if should_split else filter_item.value.split("|")
    key_node = "Word", filter_item.field
    value_node: Node
    if len(values) > 1:
        value_node = "ValueGroup", [("Word", v) for v in values if v]
    else:
        value_node = "Word", value

    if len(value_node[1]) == 0:
        raise Exception("Invalid filter arguments")

    return "KVTerm", (key_node, value_node)


def _filters_to_and_expr(filter_by: Iterable[Filter]) -> Node:
    return "AndExpression", [_filter_to_node(item) for item in filter_by]


def _apply_filters_fn(
    column_mappings: ColumnMappings,
) -> Callable[[QueryType, Iterable[Filter], CallableErrorHandler], QueryType]:
    def _apply_filters(
        stmt: QueryType, filter_by: Iterable[Filter], handle_filter_error: CallableErrorHandler
    ) -> QueryType:
        try:
            node = _filters_to_and_expr(filter_by)
            visitor = SQLAlchemyVisitor(stmt, column_mappings)
            stmt = visitor.visit_and_expression(stmt, node)
        except ValueError as exception:
            handle_filter_error(str(exception))
        return stmt

    return _apply_filters


def generic_filter_from_clauses(
    column_mappings: ColumnMappings,
) -> Callable[[QueryType, list[Filter], CallableErrorHandler], QueryType]:
    _validate_filters = generic_filters_validate(column_mappings.keys())
    _apply_filters = _apply_filters_fn(column_mappings)

    def _filter(
        query: QueryType,
        filter_by: list[Filter],
        handle_filter_error: CallableErrorHandler,
    ) -> QueryType:
        try:
            invalid_filter_items, valid_filter_items = _validate_filters(filter_by)
            if invalid_list := [item.field for item in invalid_filter_items]:
                valid_filter_keys = sorted([to_camel(k) for k in column_mappings.keys()])
                handle_filter_error(
                    "Invalid filter arguments", invalid_filters=invalid_list, valid_filter_keys=valid_filter_keys
                )

            return _apply_filters(query, valid_filter_items, handle_filter_error)
        except Exception as e:
            handle_filter_error(str(e))
            return query.where(false())

    return _filter


def create_memoized_field_list(column_mappings: dict[str, Any]) -> Callable[[], list[str]]:
    """Used to evaluate the list of keys for filtering/sorting once on first invocation.

    This is necessary to get the fully initialized list values, which can be updated during module importing.
    It works because Python closures are late-binding.
    """

    @cache
    def _filter_fields() -> list[str]:
        return sorted(column_mappings.keys())

    return _filter_fields



import contextlib
from collections.abc import Iterable, Iterator
from typing import Any
from uuid import UUID

from pydantic import BaseModel
from sqlalchemy import select

from orchestrator.db import ProductBlockTable, SubscriptionInstanceTable, db
from orchestrator.types import filter_nonetype, get_origin_and_args, is_union_type
from orchestrator.utils.functional import group_by_key
from pydantic_forms.types import UUIDstr


def _to_product_block_field_type_iterable(product_block_field_type: type | tuple[type]) -> Iterable[type]:
    """Return an iterable of types in the given product block field.

    Notes:
        There is a type-checking pattern (if optional, if union, if list/tuple, etc) which occurs multiple times in
        base.py and other parts of the codebase. We should try to combine those into one function.
        However, this function works based on field types as returned by
        DomainModel._get_depends_on_product_block_types() so it is not directly applicable.
    """
    _origin, args = get_origin_and_args(product_block_field_type)

    if is_union_type(product_block_field_type):
        return list(filter_nonetype(args))

    if isinstance(product_block_field_type, tuple):
        return product_block_field_type

    return [product_block_field_type]


@contextlib.contextmanager
def no_private_attrs(model: Any) -> Iterator:
    """PrivateAttrs from the given pydantic BaseModel are removed for the duration of this context."""
    if not isinstance(model, BaseModel):
        yield
        return
    private_attrs_reference = model.__pydantic_private__
    try:
        model.__pydantic_private__ = {}
        yield
    finally:
        model.__pydantic_private__ = private_attrs_reference


def get_root_blocks_to_instance_ids(subscription_id: UUID | UUIDstr) -> dict[str, list[UUID]]:
    """Returns mapping of root product block names to list of subscription instance ids.

    While recommended practice is to have only 1 root product block, it is possible to have multiple blocks or even a
    list of root blocks. This function supports that.
    """
    block_name_to_instance_id_rows = db.session.execute(
        select(ProductBlockTable.name, SubscriptionInstanceTable.subscription_instance_id)
        .select_from(SubscriptionInstanceTable)
        .join(ProductBlockTable)
        .where(SubscriptionInstanceTable.subscription_id == subscription_id)
        .order_by(ProductBlockTable.name)
    ).all()

    return group_by_key(block_name_to_instance_id_rows)  # type: ignore[arg-type]



# Copyright 2019-2022 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
from threading import Lock, Thread
from time import sleep, time
from typing import Optional

from structlog import get_logger

logger = get_logger(__name__)


class MemoryDistLockManager(Thread):
    """Create Distributed Locks in memory.

    This is the default implementation when not using Redis.

    Locks are limited to the current process, therefore this implementation should not be used
    in a setup with multiple gunicorn workers.
    """

    manager_lock: Lock = Lock()
    locks: dict[str, tuple[Lock, float]] = {}

    def __init__(self) -> None:
        super().__init__()
        self.daemon = True
        self.name = "MemoryDistLockExpirationThread"
        self.start()

    async def connect_redis(self) -> None:
        pass

    async def disconnect_redis(self) -> None:
        pass

    async def get_lock(
        self, resource: str, expiration_seconds: int
    ) -> Optional[Lock]:  # https://github.com/python/cpython/issues/114315
        with self.manager_lock:
            resource_lock, expire_at = self.locks.get(resource, (Lock(), time() + expiration_seconds))
            if not resource_lock.acquire(blocking=False):
                logger.debug("Resource is already locked", resource=resource, expire_at=expire_at)
                return None
            logger.debug("Successfully locked resource", resource=resource, expire_at=expire_at)
            self.locks[resource] = resource_lock, expire_at
            return resource_lock

    async def release_lock(self, lock: Lock) -> None:
        with self.manager_lock:
            lock.release()
            for resource, (resource_lock, _) in self.locks.items():
                if lock is resource_lock:
                    name = resource
                    break
            del self.locks[name]
            logger.debug("Successfully unlocked resource", resource=name)

    def release_sync(self, lock: Lock) -> None:
        asyncio.run(self.release_lock(lock))

    def run(self) -> None:
        while True:
            with self.manager_lock:
                timestamp = time()
                for resource, (resource_lock, expire_at) in list(self.locks.items()):
                    if expire_at > timestamp:
                        continue
                    resource_lock.release()
                    del self.locks[resource]
                    logger.debug("Successfully unlocked expired resource", resource=resource, expire_at=expire_at)
            sleep(0.1)


__all__ = [
    "MemoryDistLockManager",
    "Lock",
]



from sqlalchemy import inspect

from orchestrator.db.database import BaseModel
from orchestrator.db.filters.search_filters.inferred_filter import filter_exact, inferred_filter, node_to_str_val

__all__ = ["inferred_filter", "default_inferred_column_clauses", "node_to_str_val", "filter_exact"]

from orchestrator.utils.search_query import WhereCondGenerator


def default_inferred_column_clauses(table: type[BaseModel]) -> dict[str, WhereCondGenerator]:
    return {key: inferred_filter(column) for key, column in getattr(inspect(table), "columns", {}).items()}



from collections.abc import Callable

from sqlalchemy import Column
from sqlalchemy.inspection import inspect
from sqlalchemy.sql import expression

from orchestrator.db import ProductTable, SubscriptionTable
from orchestrator.db.filters import create_memoized_field_list
from orchestrator.db.sorting import QueryType, SortOrder, generic_column_sort, generic_sort
from orchestrator.utils.helpers import to_camel


def generic_subscription_relation_sort(field: Column) -> Callable[[QueryType, SortOrder], QueryType]:
    def sort_function(query: QueryType, order: SortOrder) -> QueryType:
        if order == SortOrder.DESC:
            return query.order_by(expression.desc(field))
        return query.order_by(expression.asc(field))

    return sort_function


SUBSCRIPTION_PRODUCT_SORT = {
    to_camel(key if "product" in key else f"product_{key}"): generic_subscription_relation_sort(value)
    for key, value in inspect(ProductTable).columns.items()
}

subscription_table_sort = {
    to_camel(key): generic_column_sort(value, SubscriptionTable)
    for [key, value] in inspect(SubscriptionTable).columns.items()
}

SUBSCRIPTION_SORT_FUNCTIONS_BY_COLUMN = (
    SUBSCRIPTION_PRODUCT_SORT | subscription_table_sort | {"tag": SUBSCRIPTION_PRODUCT_SORT["productTag"]}
)


subscription_sort_fields = create_memoized_field_list(SUBSCRIPTION_SORT_FUNCTIONS_BY_COLUMN)
sort_subscriptions = generic_sort(SUBSCRIPTION_SORT_FUNCTIONS_BY_COLUMN)
